#
# parser generater for ecpg
# call with backend parser as stdin
#
# Copyright (c) 2007-2008, PostgreSQL Global Development Group
#
# Written by Mike Aubury <mike.aubury@aubit.com>
#	     Michael Meskes <meskes@postgresql.org>
#
# Placed under the same license as PostgreSQL.
#

function include_stuff(includestream, includefilename, includeblock, copy) {
	copied = 0;
	while (getline < includefilename)  {
		if (includeblock != "" && $1 == "ECPG:") {
			if ($2 == includeblock) {
				copy = 1;
				includetype = $3
				if (includetype == "rule")
					dump_fields(stmt_mode, fields, field_count, " { ");
				else if (includetype == "addon")
					add_to_buffer("rules", " { ");	
			}
			else copy = 0;
		}
		else {
			if (copy == 1) {
				add_to_buffer(includestream, $0);
				copied = 1;
			}
		}
	}
	close(includefilename)
	if (includetype == "addon") {
		dump_fields(stmt_mode, fields, field_count, "");
	}
	field_count = 0;
	line = "";
	return copied;
}

function add_to_buffer(buffer, str)  {
	buff[buffer, buffcnt[buffer]++]=str
}

function dump(buffer)  {
	print "/* " buffer " */"
	for (a=0; a<buffcnt[buffer]; a++) {
		print buff[buffer,a]
	}
}

# Dump the action for a rule - 
# mode indicates if we are processing the 'stmt:' rule (mode==0 means normal,  mode==1 means stmt:)
# flds are the fields to use. These may start with a '$' - in which case they are the result of a previous non-terminal
#                             if they dont start with a '$' then they are token name
#
# len is the number of fields in flds...
# leadin is the padding to apply at the beginning (just use for formatting)
function dump_fields(mode, flds, len, ln) {

	if (mode==0) { #Normal 
		add_to_buffer("rules", ln);
		if (feature_not_supported == 1) {
			# we found an unsupported feature, but we have to
			# filter out ExecuteStmt: CREATE OptTemp TABLE ...
			# because the warning there is only valid in some situations
			if (flds[0] != "create" || flds[2] != "table")
				add_to_buffer("rules", "mmerror(PARSE_ERROR, ET_WARNING, \"unsupported feature will be passed to backend\\n\");");
			feature_not_supported = 0;
		}	
				
		if (len == 0) { # We have no fields ? 
			add_to_buffer("rules", " $$=EMPTY; }");
		} else {
			# Go through each field and try to 'aggregate' the tokens into a single 'make_str' where possible
			cnt = 0;
			for (z=0; z<len; z++) { 
				if (substr(flds[z],1,1) == "$") {
					flds_new[cnt++] = flds[z];
					continue
				}

				str = flds[z];

				while (1) {
					if (z>=len-1 || substr(flds[z+1],1,1)=="$") { # We're at the end...
						flds_new[cnt++] = "make_str(\"" str "\")";
						break;
					}
					z++;
					str = str " " flds[z];
				}
			}
			
			# So - how many fields did we end up with ? 
			if (cnt == 1) {
				# Straight assignement
				str = " $$ = " flds_new[0] ";" 
				add_to_buffer("rules", str);
			} else {
				# Need to concatenate the results to form
				# our final string
				str = " $$ = cat_str(" cnt

				for (z=0; z<cnt; z++)
					str = str "," flds_new[z]; 	

				str = str ");"
				add_to_buffer("rules",str)
			}
			if (literal_mode == 0)
				add_to_buffer("rules", "}");
		}
	} else {
		# we're in the stmt: rule
		if (len) {
			# or just the statement ...
			add_to_buffer("rules", " { output_statement($1, 0, ECPGst_normal); }");
		} else {
			add_to_buffer("rules", " { $$ = NULL; }");
		}
	}
}

function dump_line(stmt_mode, fields, field_count)  {
	block = non_term_id line;
	gsub (" ", "", block);
	gsub ("\|", "", block);
	if (rule[block] == "ignore")
		return;
	else if (rule[block])
		block = rule[block];
	if (add[block])
		line = line " " add[block];
	add_to_buffer("rules", line)
	i = include_stuff("rules", "ecpg.addons", block, 0);
	if (i == 0) dump_fields(stmt_mode, fields, field_count, " { ");
}

BEGIN {
	copymode = "off";
	brace_indent = 0;
	yaccmode = 0;
	header_included = 0;
	feature_not_supported = 0;
	tokenmode = 0;

	# some token have to be replaced by other symbols
	# either in the rule
	replace_token["BCONST"] = "ecpg_bconst"
	replace_token["FCONST"] = "ecpg_fconst"
	replace_token["IDENT"] = "ecpg_ident"
	replace_token["PARAM"] = "ecpg_param"
	# or in the block
	replace_string["WITH_TIME"] = "with time"
	replace_string["NULLS_FIRST"] = "nulls first"
	replace_string["NULLS_LAST"] = "nulls last"
	replace_string["TYPECAST"] = "::"

	# specific types for specific non-terminals - never include the ':'
	# ECPG-only types are defined in ecpg-types
	types["PrepareStmt"] = "<prep>"
 	types["opt_array_bounds"] = "<index>"
	# "ignore" means: do not create types and rules for this non-term-id
	types["stmtblock"] = "ignore"
	types["stmtmulti"] = "ignore"
	types["CreateAsStmt"] = "ignore"
	types["DeallocateStmt"] = "ignore"
	types["RuleStmt"] = "ignore"
	types["ColLabel"] = "ignore"
	types["unreserved_keyword"] = "ignore"

	# some production rules have to be ignored or replaced
	rule["stmtDeleteStmt"] = "output_statement1normal";
	rule["stmtDiscardStmt"] = "output_statement1normal";
	rule["stmtFetchStmt"] = "output_statement1normal";
	rule["stmtInsertStmt"] = "output_statement1normal";
	rule["stmtSelectStmt"] = "output_statement1normal";
	rule["stmtUpdateStmt"] = "output_statement1normal";
	rule["var_valueNumericOnly"] = "localvar1";
	rule["fetch_directionABSOLUTE_PSignedIconst"] = "localvar2";
	rule["fetch_directionRELATIVE_PSignedIconst"] = "localvar2";
	rule["fetch_directionSignedIconst"] = "localvar1";
	rule["fetch_directionFORWARDSignedIconst"] = "localvar2";
	rule["fetch_directionBACKWARDSignedIconst"] = "localvar2";
	rule["fetch_direction"] = "ignore";
	rule["IconstVarIconst"] = "civar";
	rule["SconstVarSconst"] = "civar";
	rule["var_nameColId"] = "ignore";
	rule["opt_array_boundsopt_array_bounds'['IconstVar']'"] = "ignore";
	rule["col_name_keywordCHAR_P"] = "ignore";
	rule["col_name_keywordINT_P"] = "ignore";
	rule["col_name_keywordVALUES"] = "ignore";
	rule["reserved_keywordTO"] = "ignore";
	rule["reserved_keywordUNION"] = "ignore";

	# and some stuff needs to be added
	add["VariableShowStmtSHOWvar_name"] = "ecpg_into";
	add["VariableShowStmtSHOWTIMEZONE"] = "ecpg_into";
	add["VariableShowStmtSHOWTRANSACTIONISOLATIONLEVEL"] = "ecpg_into";
	add["VariableShowStmtSHOWSESSIONAUTHORIZATION"] = "ecpg_into";
	add["ExecuteStmtEXECUTEprepared_nameexecute_param_clause"] = "execute_rest";
}

/ERRCODE_FEATURE_NOT_SUPPORTED/ {
	feature_not_supported = 1;
	next;
}

/^%%/	{
	tokenmode = 2;
	copymode = "on";
	yaccmode++;
	infield = 0;
	fieldcount = 0;
}

{
	s = $0;
	prec = 0;
	# Make sure any braces are split
	gsub("{"," { ",s);
	gsub("}"," } ",s);
	# Any comments are split
	gsub("[/][*]"," /* ",s);
	gsub("[*][/]"," */ ",s);

	# Now split the line into individual fields
	n=split(s, arr);

	if (arr[1] == "%token" && tokenmode == 0) {
		tokenmode = 1;
		include_stuff("tokens", "ecpg.tokens", "", 1);
		type = 1;
	}
	else if (arr[1] == "%type" && header_included == 0) {
		include_stuff("header", "ecpg.header", "", 1);
		include_stuff("ecpgtype", "ecpg.type", "", 1);
		header_included = 1;
	}

	if (tokenmode==1) {
		str = "";
		for (a=1; a<=n; a++) {
			if (arr[a]=="/*") {comment++; continue}
			if (arr[a]=="*/") {comment--; continue}
			if (comment) {continue;}	
			if (substr(arr[a],1,1)=="<")  continue; # its a type

			tokens[arr[a]] = 1;

			str = str " " arr[a];
			if (arr[a] == "IDENT" && arr[a-1] == "%nonassoc") 
				# add two more tokens to the list
				str = str "\n%nonassoc CSTRING\n%nonassoc UIDENT";
		}
		add_to_buffer("orig_tokens", str);
		next;
	}

	# Dont worry about anything if we're not in the right section of gram.y
	if (yaccmode!=1) {
		next;
	}

	# Go through each field in turn
	for (fieldIndexer=1; fieldIndexer<=n; fieldIndexer++) {
		if (arr[fieldIndexer]=="*/" && comment) { 
			comment = 0;
			continue;
		} else if (comment) {
			continue;
		} else if (arr[fieldIndexer]=="/*") {
			# start of a multiline comment
			comment = 1;
			continue;
		} else if (arr[fieldIndexer]=="//")
			next;
		else if (arr[fieldIndexer]=="}") {
			brace_indent--;
			continue;
		}
		else if (arr[fieldIndexer]=="{") {
			brace_indent++;	
			continue;
		}

		if (brace_indent>0) continue;

		if (arr[fieldIndexer]==";") {
			if (copymode == "on" ) {
				if (infield && includetype == "")
	                                dump_line(stmt_mode, fields, field_count);
				add_to_buffer("rules",";\n\n")
			} else 
				copymode = "on";
			field_count = 0;
			infield = 0;
			line = "";
			includetype = "";
			continue;
		}


		if (arr[fieldIndexer]=="|") {
			if (copymode == "on") {
				if (infield && includetype == "")
					dump_line(stmt_mode, fields, field_count);
				line = "| ";
			}
			field_count = 0;
			infield = 1;
			includetype = "";
			continue;
		}

		if (replace_token[arr[fieldIndexer]])
			arr[fieldIndexer] = replace_token[arr[fieldIndexer]] ;

		# Are we looking at a declaration of a non-terminal ? 
		if ((arr[fieldIndexer] ~ "[A-Za-z0-9]+:") || arr[fieldIndexer+1]==":") {
			non_term_id = arr[fieldIndexer];
			gsub(":","",non_term_id);

			if (types[non_term_id]=="") 
				types[non_term_id]="<str>";

			if (types[non_term_id]=="ignore") {
				copymode = ";";
				line = "";
				next;
			}
			else copymode = "on";

			line=line " " arr[fieldIndexer];
			# Do we have the : attached already ? 
			# If yes, we'll have already printed the ':'
			if (!(arr[fieldIndexer] ~ "[A-Za-z0-9]+:")) {
				# Consume the ':' which is next...
				line= line ":"
				fieldIndexer++;
			}

			# Special mode? 
			if (non_term_id=="stmt")
				stmt_mode=1;
			else 
				stmt_mode=0;

			tstr="%type " types[non_term_id] " " non_term_id;
			add_to_buffer("types", tstr);

			if (copymode == "on")
				add_to_buffer("rules", line);
			line = "";
			field_count = 0;
			infield = 1;
			continue;
		}
		else if (copymode == "on") line=line " " arr[fieldIndexer];

		if (arr[fieldIndexer]=="%prec") {
			prec = 1;
			continue;
		}

		if (copymode == "on" && !prec && !comment  && arr[fieldIndexer]!="/*EMPTY*/"  && length(arr[fieldIndexer]) && infield) {
			nfield = field_count+1;
			if (arr[fieldIndexer] != "Op" && (tokens[arr[fieldIndexer]]>0 || arr[fieldIndexer] ~ "'.+'") || stmt_mode == 1) {
				if (replace_string[arr[fieldIndexer]]) s = replace_string[arr[fieldIndexer]];
				else s=arr[fieldIndexer];
				gsub("_P","",s)
				gsub("'","",s)
				if (stmt_mode == 1)
					fields[field_count++] = s;
				else
					fields[field_count++] = tolower(s);
			} else {
				fields[field_count++] = "$" nfield;
			}
		}
	}
}

END {
	dump("header");
	dump("tokens");
	dump("types");
	dump("ecpgtype");
	dump("orig_tokens");
	print "%%";
	print "prog: statements;"
	dump("rules");
	include_stuff("trailer", "ecpg.trailer", "", 1);
	dump("trailer");
}
